﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Problems.BinPacking.Shapes;
using HeuristicLab.Data;
using HeuristicLab.Common;
using HeuristicLab.Problems.BinPacking.PackingPlans;

namespace HeuristicLab.Problems.BinPacking.Evaluators {
  [Item("Bin-Utilization Regular Identical-Bin Evaluator", "<Description missing> Found in Lee:2010")]
  [StorableClass]
  public abstract class BinUtilizationRegularIdenticalBinEvaluator<D, B, I> : RegularSimpleRotationIdenticalBinPackingPlanEvaluator<D, B, I>
    where D : class, IPackingDimensions
    where B : PackingShape<D>, IPackingBin, IRegularPackingShape
    where I : PackingShape<D>, IPackingItem, IRegularPackingShape {

    [StorableConstructor]
    protected BinUtilizationRegularIdenticalBinEvaluator(bool deserializing) : base(deserializing) { }
    protected BinUtilizationRegularIdenticalBinEvaluator(BinUtilizationRegularIdenticalBinEvaluator<D, B, I> original, Cloner cloner)
      : base(original, cloner) {
    }
    public BinUtilizationRegularIdenticalBinEvaluator() : base() { }

    protected override DoubleValue Evaluate() {
      DoubleValue quality = new DoubleValue(0);

      IPackingPlan plan = PackingPlanParameter.ActualValue;
      B binMeasure = PackingBinMeasuresParameter.ActualValue;
      ItemList<I> itemMeasures = PackingItemMeasuresParameter.ActualValue;
      int nrOfBins = plan.NrOfBins;


      //Check if data is valid
      //if (plan.PackingItemPositions.Count != itemMeasures.Count)
      //  throw new Exception("ERROR: ItemMeasures.count does not match packingPosition.count");


      ////Check if any items are overlapping or not contained in their bins
      //bool itemPositionsAreValid = !HasOverlappingOrNotContainedItems(plan.PackingItemPositions, binMeasure, itemMeasures, nrOfBins);



      //if (itemPositionsAreValid)
      return CalculatePackingRatio(plan as PackingPlan<D, B, I>, binMeasure, itemMeasures, nrOfBins);

      //return quality;
    }

    private DoubleValue CalculatePackingRatio(PackingPlan<D, B, I> plan, B binMeasure, ItemList<I> itemMeasures, int nrOfBins) {
      double result = 0;

      double usableSpace = binMeasure.MultipliedMeasures;
      for (int i = 0; i < nrOfBins; i++) {
        var indexes = plan.PackingItemPositions.Select((Value, Index) => new { Value, Index }).Where(s => s.Value.Value.AssignedBin == i).Select(s => s.Index);
        var packedItemsInThisBin = itemMeasures.Select((Value, Index) => new { Value, Index }).Where(s => indexes.Contains(s.Index));
        double usedSpaceInThisBin = packedItemsInThisBin.Select(s => s.Value.MultipliedMeasures).Sum();
        result += (usedSpaceInThisBin / usableSpace);
      }

      result = result / nrOfBins;
      return new DoubleValue (result);
    }
  }
}
